# -*- coding: utf-8 -*-


"""A package to calculate molecular descriptors."""


from .chemopy import ChemoPy  # noqa: F401
from .getmol import MolFrom   # noqa: F401
from .fingerprint import Fingerprint # noqa: F401
from .fingerprint3d import Fingerprint3D # noqa: F401


__version__ = '1.0.1'
